#include <stdlib.h>

#include "fompi.h"

int foMPI_Win_free(foMPI_Win *win) {

  dmapp_return_t status;
  dmapp_pe_t pe = (*win)->commrank;
  int i;
  foMPI_Win_dynamic_element_t* current;
  foMPI_Win_dynamic_element_t* previous;

  /* TODO: Could this be removed? */
  MPI_Barrier((*win)->comm);

  /* this should never be an issue, since the barrier ensure that all processes reached win_free */
  int flag;
  status = dmapp_gsync_test( &flag ); /* make sure there are no synchronization operations in the system */
  _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
  if (flag == 0) {
    status = dmapp_gsync_wait();
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
  }


  if ( (*win)->create_flavor ==  foMPI_WIN_FLAVOR_DYNAMIC ) {

    for( i = 0 ; i<(*win)->commsize ; i++ ) {

      if( i == pe ) {

        while( (*win)->win_dynamic_mem_regions != NULL ) {
        
          status = dmapp_mem_unregister(&((*win)->win_dynamic_mem_regions->seg));
          _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);

          status = dmapp_mem_unregister(&((*win)->win_dynamic_mem_regions_seg));
          _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);

          previous = (*win)->win_dynamic_mem_regions;

          (*win)->win_dynamic_mem_regions_seg = (*win)->win_dynamic_mem_regions->next_seg; /* TODO: just use pointer? */
          (*win)->win_dynamic_mem_regions = (*win)->win_dynamic_mem_regions->next;
    
          free( previous );
        }

      } else {
        if ( (*win)->dynamic_list[i].remote_mem_regions != NULL ) {
          current = (*win)->dynamic_list[i].remote_mem_regions;
          while( current->next != NULL ) {
            previous = current;
            current = current->next;
            free( previous );
          }
          free( current );
        }
      }
    }

    status = dmapp_mem_unregister(&((*win)->dynamic_list[pe].win_ptr_seg));
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);

    status = dmapp_mem_unregister(&((*win)->dynamic_list[pe].pscw_matching_exposure_seg));
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);

    free((*win)->dynamic_list[pe].pscw_matching_exposure);
    free((*win)->dynamic_list);

  } else {

#ifdef XPMEM
    /* unmap the remote memory */
    for( i=0 ; i<(*win)->onnode_size ; i++ ) {
      if ((*win)->xpmem_array[i].base_apid != -1) {
        //printf("%i: unmap base\n", debug_pe);
        foMPI_unmap_memory_xpmem( &((*win)->xpmem_array[i].base_apid), (*win)->xpmem_array[i].base, (*win)->xpmem_array[i].base_offset );
      }
      foMPI_unmap_memory_xpmem( &((*win)->xpmem_array[i].win_ptr_apid), (*win)->xpmem_array[i].win_ptr, (*win)->xpmem_array[i].win_ptr_offset );
      foMPI_unmap_memory_xpmem( &((*win)->xpmem_array[i].pscw_matching_exposure_apid), (*win)->xpmem_array[i].pscw_matching_exposure, (*win)->xpmem_array[i].pscw_matching_exposure_offset );
    }
    free( (*win)->xpmem_array );
#endif

    if ( (*win)->win_array[pe].base != NULL ) {
      status = dmapp_mem_unregister(&((*win)->win_array[pe].seg));
      _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
#ifdef XPMEM
      if ((*win)->xpmem_segdesc.base.seg != -1 ) {
        //printf("%i: unexport base\n", debug_pe);
        foMPI_unexport_memory_xpmem( (*win)->xpmem_segdesc.base );
      }
#endif
    }

    if ( (*win)->create_flavor == foMPI_WIN_FLAVOR_ALLOCATE ) {
      free( (*win)->win_array[pe].base );
    }

    status = dmapp_mem_unregister(&((*win)->win_array[pe].win_ptr_seg));
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
#ifdef XPMEM
    foMPI_unexport_memory_xpmem( (*win)->xpmem_segdesc.win_ptr );
#endif

    status = dmapp_mem_unregister(&((*win)->win_array[pe].pscw_matching_exposure_seg));
    _check_status(status, DMAPP_RC_SUCCESS,  (char*) __FILE__, __LINE__);
#ifdef XPMEM
    foMPI_unexport_memory_xpmem( (*win)->xpmem_segdesc.pscw_matching_exposure );
#endif

    free((*win)->win_array[pe].pscw_matching_exposure);
    free((*win)->win_array);

#ifdef XPMEM
    if ( (*win)->onnode_lower_bound == -1 ) {
      free( (*win)->onnode_ranks );
    }
    MPI_Group_free( &((*win)->win_onnode_group) );
    MPI_Comm_free( &((*win)->win_onnode_comm) );
#endif
  }

  free((*win)->pscw_matching_access);
  free((*win)->win_locks);
  free((*win)->group_ranks);
  if ( (*win)->name != NULL ) {
    free( (*win)->name );
  }
  MPI_Barrier( (*win)->comm );
  free(*win);

  *win = foMPI_WIN_NULL;

  return MPI_SUCCESS; 
}
